#lang racket
(require "spur.scm" (lib "defmacro.ss"))
(require (lib "readline/readline"))
(provide (all-from-out racket) (all-from-out "spur.scm")
     make-note note-onset note-pitch note-vel note-length
     set-note-onset set-note-pitch set-note-vel set-note-length
     phrase-append phrase-sequence phrase-length phrase-shift
     string->phrase schedule-phrase rppl)

;;;  This is a variant of the phrase idea from Tim Thompson's KeyKit.
;;;  It provides a convenient way of entering notes by using a string
;;;  like "C,E,G" (for a sequence of notes) or "C+E+G" (for a chord).

;; white note characters -> MIDI values (0 to 11)
(define whites (hash #\C 0 #\D 2 #\E 4 #\F 5 #\G 7 #\A 9 #\B 11))

;; digit characters -> single-digit numbers (0 to 9)
(define digits 
  (hash #\0 0 #\1 1 #\2 2 #\3 3 #\4 4 #\5 5 #\6 6 #\7 7 #\8 8 #\9 9))

;; loudness characters -> MIDI velocity (35 to 125 only)
(define louds 
  (hash "ppp" 35 "pp" 50 "p" 63 "mp" 73 "m" 80 "mf" 77 "f" 87 "ff" 
    110 "fff" 125))

;; get a number from part of a string
(define (make-number str n m) (string->number (substring str n m)))

;; get a pitch from a scale note value and an octave
(define (make-pitch scale oct) (+ (* 12 (+ oct 1)) scale))

;; test if a string index is present in a hash table
(define (string-mem str n hash)
  (and (< n (string-length str)) (hash-ref hash (string-ref str n) #f)))

;; test if a string index is equal to a char
(define (string-eq str n char)
  (and (< n (string-length str)) (char=? (string-ref str n) char)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  Operations on musical notes =  onset x len x vel x pitch
;;;  where onset is (0-*), len is (0-1023), vel and pitch are (0-127)

(define (make-note ons pit vel len) 
  (bitwise-ior (bitwise-and pit 127)
               (arithmetic-shift (bitwise-and vel 127) 7)
               (arithmetic-shift (bitwise-and len 1023) 14)
               (arithmetic-shift ons 24)))

(define (note-pitch n) (bitwise-and n 127))
(define (note-vel n) (bitwise-and (arithmetic-shift n -7) 127))
(define (note-length n) (bitwise-and (arithmetic-shift n -14) 1023))
(define (note-onset n) (arithmetic-shift n -24))
 
(define (set-note-pitch n v . rest)
  (let ((pitch (if (null? rest) v (apply v (note-pitch n) rest))))
    (bitwise-ior (bitwise-and pitch 127)
                 (arithmetic-shift (arithmetic-shift n -7) 7))))

(define (set-note-vel n v . rest)
  (define vel (real->integer (if (null? rest) v (apply v (note-vel n) rest))))
  (bitwise-ior (bitwise-and n 127)
               (arithmetic-shift (bitwise-and vel 127) 7)
               (arithmetic-shift (arithmetic-shift n -14) 14)))

(define (set-note-length n v . rest)
  (let ((length (if (null? rest) v (apply v (note-length n) rest))))
    (bitwise-ior (bitwise-and n 16383)
                 (arithmetic-shift (bitwise-and length 1023) 14)
                 (arithmetic-shift (arithmetic-shift n -24) 24))))

(define (set-note-onset n v . rest)
  (let ((onset (if (null? rest) v (apply v (note-onset n) rest))))
    (bitwise-ior (bitwise-and n 16777215)
                 (arithmetic-shift onset 24))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  Operations on musical phrases = list of notes

;; join two phrases in parallel
(define phrase-append append)

;; join two phrases in sequence
(define (phrase-sequence . phrs)
  (let loop ((shift 0) (phrs phrs))
    (if (null? phrs) '()
        (append (phrase-shift (car phrs) shift)
                (loop (+ shift (phrase-length (car phrs))) (cdr phrs)))))) 

;; the length of a phrase = the last note to end
(define (phrase-length phr) 
  (let loop ((phr phr) (m 0))
    (if (null? phr) m (loop (cdr phr) (max m (+ (note-onset (car phr))
                                                (note-length (car phr))))))))

;; add to the onset of every note in a phrase
(define (phrase-shift phr len) (map (lambda (n) (set-note-onset n + len)) phr))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  Playing musical phrases

(define default-staccato 6)                       ; 1/64th note less length
(define default-bpm 120)                          ; 120 beats per minute

;; put the notes of a phrase onto the schedule
(define (schedule-phrase time chan phr [bpm default-bpm])
  (define time-per-click (/ *minute* (* bpm default-ppq)))
  (for-each
     (lambda (n)
       (schedule-note (+ time (* (note-onset n) time-per-click)) chan
                      (note-pitch n) (note-vel n) 
                      (* (- (note-length n) default-staccato) time-per-click)))
     phr))

;; repeatedly play parsed input strings 
(define (rppl [chan 1] [bpm default-bpm])
  (display "Enter strings like \"C,D,E\" without quotes. ")
  (display "Quit with the empty string.\n")
  (with-midi-out 
    (let loop ((str (readline ">> ")))
      (when (not (string=? "" str))
        (add-history str) 
        (let ((phr (string->phrase str)))
          (if (not phr) (display "Does not parse. Quit with empty string.\n")
              (with-schedule
               (schedule-phrase (+ (now) (* .5 *second*)) chan phr bpm))))
        (loop (readline ">> ")) )))
  (display "Done playing strings.\n") )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  Parsing strings as musical phrases

(define default-ppq 96)                           ; 96 clicks = 1 beat
(define default-pitch (make-pitch 0 4))           ; C4
(define default-note #f)                          ; made using above two

(define (string->phrase str [clicks default-ppq])
  (set! default-note (make-note 0 default-pitch 70 clicks))
  (let-values ([(n phr prev) (parse-phrase* 0 str)])
    (and (= n (string-length str)) phr)))

;;; the remaining parsing functions are internal to the parser

(define (parse-phrase* n str) (parse-phrase default-note n str))

(define (parse-phrase prev n str)
  (let-values ([(n* phr* prev*) (parse-term prev n str)])
    (if (or (not phr*) (not (string-eq str n* #\,))) (values n* phr* prev*)
        (let-values ([(n** phr** prev**) (parse-phrase prev* (+ n* 1) str)])
          (values n** (and phr** (phrase-sequence phr* phr**)) prev**)))))

(define (parse-term prev n str)
  (let-values ([(n* phr* prev*) (parse-factor prev n str)])
     (if (or (not phr*) (not (string-eq str n* #\+))) (values n* phr* prev*)
         (let-values ([(n** phr** prev**) (parse-term prev* (+ n* 1) str)])
           (values n** (and phr** (phrase-append phr* phr**)) prev**)))))

(define (parse-factor prev n str)
  (cond 
   ((string-eq str n #\() 
    (let-values ([(n* phr* prev*) (parse-phrase* (+ n 1) str)])
      (values (+ n* 1) (and (string-eq str n* #\)) phr*) prev*)))
   ((string-eq str n #\R) ; for rests, make a node with velocity 0
    (let-values ([(n* phr* prev*) (parse-oct 0 prev (+ n 1) str)])
      (values n* (list (set-note-pitch (set-note-vel prev* 0) 0)) prev*)))
   (else (parse-note prev n str))))

(define (parse-note prev n str)
  (if (string-eq str n #\~)
      (let-values ([(n* phr* prev*) (parse-note prev (+ n 1) str)])
        (if (null? phr*) (values n #f prev*)
            (values n* (append (trillify (car phr*)) phr*) prev*)))
      (parse-white prev n str)))

(define (trillify note)
  (let* ((this (note-pitch note)) (ons (note-onset note))
         (add (if (memq (modulo this 12) '(4 11)) 1 2)))
    (list (make-note (- ons 24) this 52 12) 
          (make-note (- ons 12) (+ this add) 55 12))))

(define (parse-white prev n str)
  (if (string-mem str n whites)
      (parse-dec (hash-ref whites (string-ref str n)) prev (+ n 1) str)
      (values n #f prev)))

(define (parse-dec white prev n str)
  (if (string-eq str n #\b) (parse-oct (- white 1) prev (+ n 1) str)
      (if (string-eq str n #\#) (parse-oct (+ white 1) prev (+ n 1) str)
          (parse-oct white prev n str))))

(define (parse-oct white prev n str)
  (if (string-mem str n digits)
      (parse-loud (make-pitch white (hash-ref digits (string-ref str n)))
                  prev (+ n 1) str)
      (parse-loud (make-pitch white (- (quotient (note-pitch prev) 12) 1))
                  prev n str)))

(define (parse-loud pitch prev n str)
  (cond  
   ((string-eq str n #\p)
    (if (string-eq str (+ n 1) #\p)
        (if (string-eq str (+ n 2) #\p)
            (parse-len pitch (hash-ref louds "ppp") prev (+ n 3) str)
            (parse-len pitch (hash-ref louds "pp") prev (+ n 2) str))
        (parse-len pitch (hash-ref louds "p") prev (+ n 1) str)))
   ((string-eq str n #\f)
    (if (string-eq str (+ n 1) #\f)
        (if (string-eq str (+ n 2) #\f)
            (parse-len pitch (hash-ref louds "fff") prev (+ n 3) str)
            (parse-len pitch (hash-ref louds "ff") prev (+ n 2) str))
        (parse-len pitch (hash-ref louds "f") prev (+ n 1) str)))
   ((string-eq str n #\m)
    (cond 
     ((string-eq str (+ n 1) #\p)
      (parse-len pitch (hash-ref louds "mp") prev (+ n 2) str))
     ((string-eq str (+ n 1) #\f)
      (parse-len pitch (hash-ref louds "mf") prev (+ n 2) str))
     (else (parse-len pitch (hash-ref louds "m") prev (+ n 1) str))))
   (else (parse-len pitch (note-vel prev) prev n str))))

(define (parse-len pitch vel prev n str)
  (if (string-eq str n #\:)
      (let loop ((m (+ n 1)))
        (if (and (< m (string-length str))
                 (hash-ref digits (string-ref str m) #f))
            (loop (+ m 1))
            (let ((note (make-note 0 pitch vel (make-number str (+ n 1) m))))
              (values m (list note) note))))
      (let ((note (make-note 0 pitch vel (note-length prev))))
        (values n (list note) note))))
